# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

import contextlib
import inspect
import logging

from unittest.mock import AsyncMock, Mock

import pytest

import svcs

from .fake_factories import (
    async_int_factory,
    async_str_cleanup_factory,
    int_factory,
    str_cleanup_factory,
)
from .ifaces import AnotherService, Interface, Service


needs_working_async_mock = pytest.mark.skipif(
    not inspect.iscoroutinefunction(AsyncMock()),
    reason="AsyncMock not working",
)


class TestRegistry:
    def test_repr_empty(self, registry):
        """
        repr of an empty registry says 0 registered services.
        """
        assert "<svcs.Registry(num_services=0)>" == repr(registry)

    def test_repr_counts(self, registry):
        """
        repr counts 2 registered services as 2.
        """
        registry.register_factory(Service, Service)
        registry.register_factory(AnotherService, AnotherService)

        assert "<svcs.Registry(num_services=2)>" == repr(registry)

    def test_empty_close(self):
        """
        Closing an empty registry does nothing.
        """
        svcs.Registry().close()

        with svcs.Registry():
            ...

    def test_close_closes(self, registry):
        """
        Calling close on Registry runs all on_close callbacks.
        """
        close_1 = Mock()
        close_2 = Mock()

        registry.register_factory(Service, Service, on_registry_close=close_1)
        registry.register_value(
            AnotherService, AnotherService, on_registry_close=close_2
        )

        registry.close()

        assert close_1.called
        assert close_2.called
        assert not registry._services
        assert not registry._on_close

    def test_overwritten_factories_are_not_forgotten(self, registry):
        """
        If a factory is overwritten, it's close callback is still called.
        """
        close_1 = Mock()
        close_2 = Mock()

        registry.register_factory(Service, Service, on_registry_close=close_1)
        registry.register_value(
            Service, AnotherService, on_registry_close=close_2
        )

        registry.close()

        assert close_1.called
        assert close_2.called

    def test_close_warns_about_async(self, registry):
        """
        Calling close raises a warning if there are async cleanups.
        """

        async def callback():
            ...

        registry.register_factory(Service, Service, on_registry_close=callback)

        with pytest.warns(
            UserWarning,
            match="Skipped async cleanup for 'tests.ifaces.Service'.",
        ):
            registry.close()

    def test_register_factory_logs(self, registry, caplog):
        """
        register_factory logs the registration to debug.
        """
        caplog.set_level(logging.DEBUG)

        registry.register_factory(Interface, Service)

        assert [
            "registered factory <class 'tests.ifaces.Service'> for "
            "service type tests.ifaces.Interface"
        ] == caplog.messages
        assert "tests.ifaces.Interface" == caplog.records[0].svcs_service_name
        assert "tests.ifaces.Service" == caplog.records[0].svcs_factory_name

    def test_register_value_logs(self, registry, caplog):
        """
        register_value logs the registration to debug.
        """
        caplog.set_level(logging.DEBUG)

        registry.register_value(Service, 42)

        assert [
            "registered value 42 for service type tests.ifaces.Service"
        ] == caplog.messages
        assert "tests.ifaces.Service" == caplog.records[0].svcs_service_name
        assert 42 == caplog.records[0].svcs_value

    def test_close_logs_failures(self, registry, caplog):
        """
        Closing failures are logged but ignored.
        """
        registry.register_factory(
            Service, Service, on_registry_close=Mock(side_effect=ValueError())
        )

        with registry:
            ...

        assert [
            "Registry's on_registry_close callback failed for 'tests.ifaces.Service'."
        ] == caplog.messages
        assert "tests.ifaces.Service" == caplog.records[0].svcs_service_name

    def test_context_manager(self):
        """
        The registry is also a context manager that closes on exit.
        """
        orc = Mock()

        with svcs.Registry() as registry:
            registry.register_factory(Service, Service, on_registry_close=orc)

        orc.assert_called_once_with()

    @pytest.mark.asyncio()
    async def test_async_context_manager(self):
        """
        The registry is also an async context manager that acloses on exit.

        Passing an awaitable to on_registry_close works.
        """
        closed = False

        async def closer():
            nonlocal closed
            closed = True

        async with svcs.Registry() as registry:
            registry.register_factory(
                Service, Service, on_registry_close=closer()
            )

        assert closed

    def test_detects_async_factories(self, registry):
        """
        The is_async property of the RegisteredService is True if the factory
        needs to be awaited.
        """

        registry.register_factory(int, async_int_factory)
        registry.register_factory(str, async_str_cleanup_factory)

        assert (
            svcs.RegisteredService(int, async_int_factory, False, True, None)
            == registry._services[int]
        )
        assert (
            svcs.RegisteredService(
                str, async_str_cleanup_factory, False, True, None
            )
            == registry._services[str]
        )

    def test_no_false_positive_async(self, registry):
        """
        is_async is False for sync factories.
        """
        registry.register_factory(int, int_factory)
        registry.register_factory(str, str_cleanup_factory)

        assert (
            svcs.RegisteredService(int, int_factory, False, False, None)
            == registry._services[int]
        )
        assert (
            svcs.RegisteredService(
                str, str_cleanup_factory, False, False, None
            )
            == registry._services[str]
        )

    @pytest.mark.skipif(
        not hasattr(contextlib, "aclosing"),
        reason="Hasn't contextlib.aclosing()",
    )
    @pytest.mark.asyncio()
    async def test_async_empty_close(self, registry):
        """
        Asynchronously closing an empty registry does nothing.
        """
        await registry.aclose()

        async with svcs.Registry():
            ...

    @pytest.mark.asyncio()
    @needs_working_async_mock
    async def test_aclose_mixed(self, registry):
        """
        aclose() closes all services, including async ones.
        """
        sync_close = Mock()
        async_close = AsyncMock()

        registry.register_factory(
            Service, Service, on_registry_close=sync_close
        )
        registry.register_factory(
            AnotherService, AnotherService, on_registry_close=async_close
        )

        await registry.aclose()

        assert sync_close.called

        async_close.assert_awaited_once()

    @pytest.mark.asyncio()
    @needs_working_async_mock
    async def test_aclose_logs_failures(self, registry, caplog):
        """
        Async closing failures are logged but ignored.
        """
        close_mock = AsyncMock(side_effect=ValueError())

        registry.register_factory(
            Service,
            Service,
            on_registry_close=close_mock,
        )

        await registry.aclose()

        close_mock.assert_awaited_once()
        assert "tests.ifaces.Service" == caplog.records[0].svcs_service_name

    def test_contains(self, registry):
        """
        If a service is registered with a registry, `in` returns True, False
        otherwise.
        """
        registry.register_factory(Service, Service)

        assert Service in registry
        assert AnotherService not in registry


class TestRegisteredService:
    def test_repr(self, rs):
        """
        repr uses the fully-qualified name of a svc type.
        """

        assert (
            "<RegisteredService(svc_type=tests.ifaces.Service, "
            "<class 'tests.ifaces.Service'>, takes_container=False, "
            "has_ping=False"
            ")>"
        ) == repr(rs)

    def test_name(self, rs):
        """
        The name property deducts the correct class name.
        """

        assert "tests.ifaces.Service" == rs.name


def wrong_annotation(foo: svcs.Registry) -> int:
    ...


def no_args():
    ...


def diff_name():
    ...


class TestTakesContainer:
    @pytest.mark.parametrize(
        "factory",
        [no_args, diff_name, wrong_annotation],
    )
    def test_nope(self, factory):
        """
        Functions with different names and annotations are ignored.
        """
        assert not svcs._core._takes_container(factory)

    def test_name(self):
        """
        Return True if the name is `svcs_container`.
        """

        def factory(svcs_container):
            ...

        assert svcs._core._takes_container(factory)

    def test_annotation(self):
        """
        Return true if the first argument is annotated as `svcs.Container`.
        """

        def factory(foo: svcs.Container):
            ...

        assert svcs._core._takes_container(factory)

    def test_annotation_str(self):
        """
        Return true if the first argument is annotated as `svcs.Container`
        using a string.
        """

        def factory(bar: "svcs.Container"):
            ...

        assert svcs._core._takes_container(factory)

    def test_catches_invalid_sigs(self):
        """
        If the factory takes more than one parameter, raise an TypeError.
        """

        def factory(foo, bar):
            ...

        with pytest.raises(
            TypeError, match="Factories must take 0 or 1 parameters."
        ):
            svcs._core._takes_container(factory)

    def test_call_works(self):
        """
        Does not raise if the factory is a class with __call__.
        """

        class Factory:
            def __call__(self, svcs_container):
                ...

        assert svcs._core._takes_container(Factory())

    def test_signature_fails(self):
        """
        If getting the signature fails, it means the factory doesn't get a
        container.
        """
        assert not svcs._core._takes_container(int)


class TestFullName:
    def test_object(self):
        """
        Object has neither ``__module__`` nor ``__qualname__``, so it gets
        repr'ed.
        """
        assert svcs._core._full_name(object()).startswith(
            "<object object at 0x"
        )



from __future__ import annotations

from aiohttp.web import Request, Response, json_response

import svcs


async def healthy_view(request: Request) -> Response:
    ok: list[str] = []
    failing: list[dict[str, str]] = []
    code = 200

    for svc in svcs.aiohttp.get_pings(request):
        try:
            await svc.aping()
            ok.append(svc.name)
        except Exception as e:
            failing.append({svc.name: repr(e)})
            code = 500

    return json_response({"ok": ok, "failing": failing}, status=code)



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from typing import Generator

import flask

import svcs


def factory_with_cleanup() -> Generator[int, None, None]:
    yield 1


reg = svcs.Registry()

app = flask.Flask("tests")
app = svcs.flask.init_app(app, registry=reg)
app = svcs.flask.init_app(app)

svcs.flask.replace_value(int, 1)
svcs.flask.replace_value(int, 1, ping=lambda: None)

svcs.flask.register_value(app, int, 1)
svcs.flask.register_value(app, int, 1, ping=lambda: None)

svcs.flask.replace_factory(str, str)
svcs.flask.replace_value(str, str, ping=lambda: None)

svcs.flask.register_factory(app, str, str)
svcs.flask.register_factory(app, int, factory_with_cleanup)
svcs.flask.register_value(app, str, str, ping=lambda: None)

o1: object = svcs.flask.get(object)

a: int
b: str
c: bool
d: tuple
e: object
f: float
g: list
h: dict
i: set
j: bytes
a, b, c, d, e, f, g, h, i, j = svcs.flask.get(
    int, str, bool, tuple, object, float, list, dict, set, bytes
)


svcs.flask.close_registry(app)

con: svcs.Container = svcs.flask.svcs_from()
con = svcs.flask.svcs_from(flask.g)


class CustomApp(flask.Flask):
    pass


app = svcs.flask.init_app(CustomApp("tests"))



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from importlib import metadata


# We want an image in the README and include the README in the docs.
suppress_warnings = ["image.nonlocal_uri"]


# -- General configuration ----------------------------------------------------

extensions = [
    "myst_parser",
    "notfound.extension",
    "sphinx.ext.autodoc",
    "sphinx.ext.autodoc.typehints",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx_inline_tabs",
]

myst_enable_extensions = [
    "colon_fence",
    "smartquotes",
    "deflist",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = [".rst", ".md"]

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "svcs"
author = "Hynek Schlawack"
copyright = f"2023, { author }"


# The full version, including alpha/beta/rc tags.
release = metadata.version("svcs")
# The short X.Y version.
version = release.rsplit(".", 1)[0]

if "dev" in release:
    release = version = "UNRELEASED"

exclude_patterns = ["_build"]

nitpick_ignore = [
    ("py:class", "aiohttp.web_request.Request"),
]

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).

# Move type hints into the description block, instead of the func definition.
autodoc_typehints = "description"
autodoc_typehints_description_target = "documented"

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"
# None of the options work, so we disable the button completely.
html_theme_options = {
    "top_of_page_button": None,
    "sidebar_hide_name": True,
}
html_logo = "_static/logo_with_name.svg"
html_static_path = ["_static"]

htmlhelp_basename = "svcsdoc"

_descr = f"{metadata.metadata('svcs')['summary']} for Python."
_title = "svcs"
rst_epilog = f"""\
.. meta::
    :property=og:type: website
    :property=og:site_name: { _title }
    :property=og:description: { _descr }
    :property=og:author: Hynek Schlawack
    :twitter:title: { _title }
    :twitter:creator: @hynek
"""

# GitHub has rate limits
linkcheck_ignore = [
    r"https://github.com/.*/(issues|pull|compare)/\d+",
    r"https://twitter.com/.*",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "aiohttp": ("https://docs.aiohttp.org/en/stable/", None),
    "flask": ("https://flask.palletsprojects.com/en/latest/", None),
    "pyramid": (
        "https://docs.pylonsproject.org/projects/pyramid/en/main/",
        None,
    ),
}



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Callable
from typing import Any, Protocol, TypeVar, overload

import attrs

from pyramid.config import Configurator
from pyramid.registry import Registry
from pyramid.request import Request
from pyramid.response import Response
from pyramid.threadlocal import get_current_registry, get_current_request

import svcs

from ._core import _KEY_CONTAINER, _KEY_REGISTRY


def svcs_from(request: Request | None = None) -> svcs.Container:
    """
    Get the current container either from *request* or from thread locals.

    Arguments:

        request: If None, thread locals are used.
    """
    if request is None:
        request = get_current_request()

    return getattr(request, _KEY_CONTAINER)  # type: ignore[no-any-return]


def init(
    config: Configurator,
    *,
    registry: svcs.Registry | None = None,
    tween_under: Any = None,
    tween_over: Any = None,
) -> None:
    """
    Configure *config* to work with *svcs*.

    *svcs* uses a :term:`tween` to manage the life cycle of the container. You
    can affect its position by passing *tween_under* and *tween_over*.

    .. _Tween: https://docs.pylonsproject.org/projects/pyramid/en/main/glossary.html#term-tween

    Args:
        config: Pyramid configurator object.

        registry: A custom *svcs* registry to use. If None, a new one is created.

        tween_under: Passed unchanged to
            :meth:`pyramid.config.Configurator.add_tween()` as *under*.

        tween_over: Passed unchanged to
            :meth:`pyramid.config.Configurator.add_tween()` as *over*.
    """
    config.registry[_KEY_REGISTRY] = registry or svcs.Registry()

    config.add_tween(
        "svcs.pyramid.ServicesTween", over=tween_over, under=tween_under
    )


@attrs.define
class ServicesTween:
    """
    Handle *svcs* container life cycle for a Pyramid request.
    """

    handler: Callable[[Request], Response]
    registry: Registry

    def __call__(self, request: Request) -> Response:
        def make_container(request: Request) -> svcs.Container:
            con = svcs.Container(self.registry[_KEY_REGISTRY])
            request.add_finished_callback(lambda _: con.close())

            return con

        request.set_property(make_container, _KEY_CONTAINER, reify=True)
        return self.handler(request)


def register_factory(
    config: PyramidRegistryHaver,
    svc_type: type,
    factory: Callable,
    *,
    ping: Callable | None = None,
    on_registry_close: Callable | None = None,
) -> None:
    """
    Same as :meth:`svcs.Registry.register_factory()`, but uses registry on
    *config*.
    """
    config.registry[_KEY_REGISTRY].register_factory(
        svc_type, factory, ping=ping, on_registry_close=on_registry_close
    )


def register_value(
    config: PyramidRegistryHaver,
    svc_type: type,
    value: object,
    *,
    ping: Callable | None = None,
    on_registry_close: Callable | None = None,
) -> None:
    """
    Same as :meth:`svcs.Registry.register_value()`, but uses registry on
    *config*.
    """
    config.registry[_KEY_REGISTRY].register_value(
        svc_type, value, ping=ping, on_registry_close=on_registry_close
    )


def close_registry(rh: PyramidRegistryHaver) -> None:
    """
    Close the registry on *rh*, if present.

    Ideal for :func:`atexit.register()` handlers.

    Parameters:
        rh: An object that carries a :class:`pyramid.registry.Registry`.
    """
    if reg := rh.registry.pop(_KEY_REGISTRY, None):
        reg.close()


class PyramidRegistryHaver(Protocol):
    """
    An object with a :class:`pyramid.registry.Registry` as a ``registry``
    attribute. For example a :class:`~pyramid.config.Configurator` or an
    application.
    """

    registry: dict[str, Any]


def get_registry(rh: PyramidRegistryHaver | None = None) -> svcs.Registry:
    """
    Get the registry from *rh* or thread locals.

    Arguments:
        rh: If None, thread locals are used.
    """
    if rh:
        return rh.registry[_KEY_REGISTRY]  # type: ignore[no-any-return]

    return get_current_registry()[_KEY_REGISTRY]  # type: ignore[no-any-return]


def get_abstract(*svc_types: type) -> Any:
    """
    Same as :meth:`svcs.Container.get_abstract()`, but uses container on
    thread locals.
    """
    return get(*svc_types)


T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")
T7 = TypeVar("T7")
T8 = TypeVar("T8")
T9 = TypeVar("T9")
T10 = TypeVar("T10")


@overload
def get(svc_type: type[T1], /) -> T1:
    ...


@overload
def get(svc_type1: type[T1], svc_type2: type[T2], /) -> tuple[T1, T2]:
    ...


@overload
def get(
    svc_type1: type[T1], svc_type2: type[T2], svc_type3: type[T3], /
) -> tuple[T1, T2, T3]:
    ...


@overload
def get(
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    /,
) -> tuple[T1, T2, T3, T4]:
    ...


@overload
def get(
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    /,
) -> tuple[T1, T2, T3, T4, T5]:
    ...


@overload
def get(
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6]:
    ...


@overload
def get(
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    svc_type7: type[T7],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6, T7]:
    ...


@overload
def get(
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    svc_type7: type[T7],
    svc_type8: type[T8],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8]:
    ...


@overload
def get(
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    svc_type7: type[T7],
    svc_type8: type[T8],
    svc_type9: type[T9],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8, T9]:
    ...


@overload
def get(
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    svc_type7: type[T7],
    svc_type8: type[T8],
    svc_type9: type[T9],
    svc_type10: type[T10],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10]:
    ...


def get(*svc_types: type) -> object:
    """
    Same as :meth:`svcs.Container.get()`, but uses thread locals to find the
    current request.
    """
    return svcs_from().get(*svc_types)



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

"""
Interfaces used throughout the tests. They're dataclasses so they have a
predicatable repr.
"""

import dataclasses

from typing import Protocol, runtime_checkable


@runtime_checkable
class Interface(Protocol):
    pass


@dataclasses.dataclass
class Service:
    pass


@dataclasses.dataclass
class AnotherService:
    pass


@dataclasses.dataclass
class YetAnotherService:
    pass



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Callable
from typing import Any, TypeVar, overload

from flask import Flask, current_app, g, has_app_context
from flask.ctx import _AppCtxGlobals

from ._core import (
    _KEY_CONTAINER,
    _KEY_REGISTRY,
    T1,
    T2,
    T3,
    T4,
    T5,
    T6,
    T7,
    T8,
    T9,
    T10,
    Container,
    Registry,
    ServicePing,
)


def svcs_from(g: _AppCtxGlobals = g) -> Container:
    """
    Get the current container from *g*.
    """
    if (con := g.get(_KEY_CONTAINER, None)) is None:
        con = Container(current_app.config[_KEY_REGISTRY])
        setattr(g, _KEY_CONTAINER, con)

    return con  # type: ignore[no-any-return]


FlaskAppT = TypeVar("FlaskAppT", bound=Flask)


def init_app(app: FlaskAppT, *, registry: Registry | None = None) -> FlaskAppT:
    """
    Initialize *app* for *svcs*.

    Creates a registry for you if you don't provide one.
    """
    app.config[_KEY_REGISTRY] = registry or Registry()
    app.teardown_appcontext(teardown)

    return app


def get_abstract(*svc_types: type) -> Any:
    """
    Same as :meth:`svcs.Container.get_abstract()`, but uses container on
    :obj:`flask.g`.
    """
    return get(*svc_types)


def register_factory(
    app: Flask,
    svc_type: type,
    factory: Callable,
    *,
    ping: Callable | None = None,
    on_registry_close: Callable | None = None,
) -> None:
    """
    Same as :meth:`svcs.Registry.register_factory()`, but uses registry on
    *app* that has been put there by :func:`init_app()`.
    """
    app.config[_KEY_REGISTRY].register_factory(
        svc_type, factory, ping=ping, on_registry_close=on_registry_close
    )


def register_value(
    app: Flask,
    svc_type: type,
    value: object,
    *,
    ping: Callable | None = None,
    on_registry_close: Callable | None = None,
) -> None:
    """
    Same as :meth:`svcs.Registry.register_value()`, but uses registry on *app*
    that has been put there by :func:`init_app()`.
    """
    app.config[_KEY_REGISTRY].register_value(
        svc_type, value, ping=ping, on_registry_close=on_registry_close
    )


def replace_factory(
    svc_type: type,
    factory: Callable,
    *,
    ping: Callable | None = None,
    on_registry_close: Callable | None = None,
) -> None:
    """
    Register *factory* for *svc_type* and clear any cached values for it in the
    current container on :obj:`flask.g`.

    .. seealso::
        :ref:`flask-testing`
    """
    container = svcs_from(g)
    registry = container.registry

    container.forget_about(svc_type)
    registry.register_factory(
        svc_type, factory, ping=ping, on_registry_close=on_registry_close
    )


def replace_value(
    svc_type: type,
    value: object,
    *,
    ping: Callable | None = None,
    on_registry_close: Callable | None = None,
) -> None:
    """
    Register *instance* for *svc_type* and clear any cached values for it.

    .. seealso::
        :ref:`flask-testing`
    """
    container = svcs_from(g)
    registry = container.registry

    container.forget_about(svc_type)
    registry.register_value(
        svc_type, value, ping=ping, on_registry_close=on_registry_close
    )


def get_pings() -> list[ServicePing]:
    """
    See :meth:`svcs.Container.get_pings()`.

    .. seealso::
        :ref:`flask-health`
    """
    return svcs_from(g).get_pings()


def teardown(exc: BaseException | None) -> None:
    """
    To be used with :meth:`flask.Flask.teardown_appcontext` that requires to
    take an exception.

    The app context is torn down after the response is sent.
    """
    if has_app_context() and (container := g.pop(_KEY_CONTAINER, None)):
        container.close()


def close_registry(app: Flask) -> None:
    """
    Close the registry on *app*, if present.
    """
    if reg := app.config.pop(_KEY_REGISTRY, None):
        reg.close()


@overload
def get(svc_type: type[T1], /) -> T1:
    ...


@overload
def get(svc_type1: type[T1], svc_type2: type[T2], /) -> tuple[T1, T2]:
    ...


@overload
def get(
    svc_type1: type[T1], svc_type2: type[T2], svc_type3: type[T3], /
) -> tuple[T1, T2, T3]:
    ...


@overload
def get(
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    /,
) -> tuple[T1, T2, T3, T4]:
    ...


@overload
def get(
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    /,
) -> tuple[T1, T2, T3, T4, T5]:
    ...


@overload
def get(
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6]:
    ...


@overload
def get(
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    svc_type7: type[T7],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6, T7]:
    ...


@overload
def get(
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    svc_type7: type[T7],
    svc_type8: type[T8],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8]:
    ...


@overload
def get(
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    svc_type7: type[T7],
    svc_type8: type[T8],
    svc_type9: type[T9],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8, T9]:
    ...


@overload
def get(
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    svc_type7: type[T7],
    svc_type8: type[T8],
    svc_type9: type[T9],
    svc_type10: type[T10],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10]:
    ...


def get(*svc_types: type) -> object:
    """
    Same as :meth:`svcs.Container.get()`, but uses container on :obj:`flask.g`.
    """
    return svcs_from(g).get(*svc_types)



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from . import exceptions
from ._core import (
    Container,
    RegisteredService,
    Registry,
    ServicePing,
)


__all__ = [
    "Container",
    "RegisteredService",
    "Registry",
    "ServicePing",
    "exceptions",
]

try:
    from . import aiohttp
except ImportError:
    __all__ += ["aiohttp"]

try:
    from . import flask
except ImportError:
    __all__ += ["flask"]

try:
    from . import pyramid
except ImportError:
    __all__ += ["pyramid"]


# Make nicer public names.
__locals = locals()
for __name in __all__:
    if not __name.startswith("__") and not __name.islower():
        __locals[__name].__module__ = "svcs"
del __locals
del __name  # pyright: ignore[reportUnboundVariable]



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import json

import attrs
import pytest

import svcs

from tests.fake_factories import async_int_factory
from tests.ifaces import Service


try:
    from aiohttp import ClientSession
    from aiohttp.web import (
        Application,
        AppRunner,
        Request,
        Response,
        TCPSite,
        json_response,
    )
    from yarl import URL
except ImportError:
    pytest.skip("AIOHTTP not installed", allow_module_level=True)


@attrs.define
class AppServer:
    """
    An HTTP server that serves an aiohttp app.
    """

    app: Application
    runner: AppRunner
    base_url: URL

    @classmethod
    async def start(cls, app: Application) -> AppServer:
        runner = AppRunner(app)
        await runner.setup()
        site = TCPSite(runner, "127.0.0.1", 0)
        await site.start()

        host, port = runner.addresses[0][:2]

        return cls(
            app=app,
            runner=runner,
            base_url=URL.build(scheme="http", host=host, port=port),
        )

    async def aclose(self):
        await self.runner.cleanup()


async def get(url: URL) -> tuple[Response, str]:
    async with ClientSession() as session, session.get(url) as resp:
        return resp, await resp.text()


async def view(request: Request) -> Response:
    return json_response(
        {
            "value": await svcs.aiohttp.aget(request, str),
            "factory": await svcs.aiohttp.svcs_from(request).aget(int),
        }
    )


async def health_view(request: Request) -> Response:
    return json_response(
        {p.name: (await p.aping()) for p in svcs.aiohttp.get_pings(request)}
    )


@pytest.fixture(name="app")
def _app(registry):
    return svcs.aiohttp.init_app(Application(), registry=registry)


@pytest.mark.asyncio()
class TestAIOHTTP:
    async def test_aclose_registry_ok(self, app):
        """
        aclose_registry closes the registry. Automatically as part of aiohttp's
        cleanup.
        """
        closed = False

        async def closer():
            nonlocal closed
            closed = True

        svcs.aiohttp.register_factory(
            app, Service, Service, on_registry_close=closer
        )

        server = await AppServer.start(app)
        await server.aclose()

        assert closed

    async def test_aclose_registry_robust(self):
        """
        aclose_registry handles lack of of registry gracefully.
        """
        await svcs.aiohttp.aclose_registry(Application())

    async def test_registrations(self, app):
        """
        Registered values are returned.
        """
        app.router.add_get("/", view)

        svcs.aiohttp.register_value(app, str, "hello, world")
        svcs.aiohttp.register_factory(app, int, async_int_factory)

        server = await AppServer.start(app)
        resp, text = await get(server.base_url)

        await server.aclose()

        assert {"factory": 42, "value": "hello, world"} == json.loads(text)

    async def test_get_registry(self, registry, app):
        """
        get_registry returns the registry from the app that is passed.
        """
        assert registry is svcs.aiohttp.get_registry(app)

    async def test_get_pings(self, registry, container, app):
        """
        Our get_pings returns registered pings.
        """
        app.router.add_get("/", health_view)

        async def aping(_):
            ...

        svcs.aiohttp.register_factory(app, int, async_int_factory, ping=aping)

        server = await AppServer.start(app)
        resp, text = await get(server.base_url)

        assert {"builtins.int": None} == json.loads(text)

        await server.aclose()



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT


import asyncio


def nop(*_, **__):
    pass


def int_factory():
    return 42


def str_cleanup_factory():
    yield "foo"


async def async_int_factory():
    await asyncio.sleep(0)
    return 42


async def async_str_cleanup_factory():
    await asyncio.sleep(0)
    yield str(42)
    await asyncio.sleep(0)



from __future__ import annotations

import flask

import svcs


bp = flask.Blueprint("instrumentation", __name__)


@bp.get("healthy")
def healthy() -> flask.ResponseValue:
    """
    Ping all external services.
    """
    ok: list[str] = []
    failing: list[dict[str, str]] = []
    code = 200

    for svc in svcs.flask.get_pings():
        try:
            svc.ping()
            ok.append(svc.name)
        except Exception as e:
            failing.append({svc.name: repr(e)})
            code = 500

    return {"ok": ok, "failing": failing}, code



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations


class ServiceNotFoundError(Exception):
    """
    Raised when a service type is not registered.
    """



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

import contextlib
import sys

from typing import AsyncGenerator, Generator, Protocol

import svcs


reg = svcs.Registry()
con = svcs.Container(reg)

reg.close()
with contextlib.closing(reg) as reg:
    ...
with reg as reg:
    reg.register_factory(int, int)


async def func() -> None:
    await reg.aclose()
    await con.aclose()

    async with svcs.Registry() as reg2:
        reg2.register_factory(int, int)

        async with svcs.Container(reg2) as con2:
            a: int
            b: str
            c: bool
            d: tuple
            e: object
            f: float
            g: list
            h: dict
            i: set
            j: bytes
            a, b, c, d, e, f, g, h, i, j = await con2.aget(
                int, str, bool, tuple, object, float, list, dict, set, bytes
            )


def gen() -> Generator:
    yield 42


async def async_gen() -> AsyncGenerator:
    yield 42


def factory_with_cleanup() -> Generator[int, None, None]:
    yield 1


def factory_that_takes_container_by_annotation(foo: svcs.Container) -> int:
    return 1


async def async_ping() -> None:
    pass


reg.register_value(int, 1)
reg.register_value(int, 1, ping=lambda: None)
reg.register_value(int, 1, ping=async_ping)
reg.register_value(int, gen)

reg.register_factory(str, str)
reg.register_factory(int, factory_with_cleanup)
reg.register_factory(int, factory_with_cleanup, ping=async_ping)
reg.register_value(str, str, ping=lambda: None)
reg.register_value(str, async_gen)

con = svcs.Container(reg)

# The type checker believes whatever we tell it.
o1: object = con.get(object)
o2: int = con.get(int)

a: int
b: str
c: bool
d: tuple
e: object
f: float
g: list
h: dict
i: set
j: bytes
a, b, c, d, e, f, g, h, i, j = con.get(
    int, str, bool, tuple, object, float, list, dict, set, bytes
)


class P(Protocol):
    def m(self) -> None:
        ...


p: P = con.get_abstract(P)

con.close()

with contextlib.closing(svcs.Container(reg)) as con:
    ...

with svcs.Container(reg) as con:
    i2: int = con.get(int)

if sys.version_info >= (3, 10):

    async def ctx() -> None:
        async with contextlib.aclosing(svcs.Container(reg)):
            ...

        async with contextlib.aclosing(svcs.Registry()):
            ...



import logging.config

from datetime import datetime

import svcs


logging.config.dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "standard": {
                "format": "%(name)s: %(message)s",
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "DEBUG",
                "formatter": "standard",
            },
        },
        "loggers": {
            "svcs": {
                "handlers": ["console"],
                "level": "DEBUG",
                "propagate": False,
            },
        },
        "root": {
            "handlers": ["console"],
            "level": "INFO",
        },
    }
)


reg = svcs.Registry()

reg.register_factory(datetime, datetime.now)
reg.register_value(str, "Hello World")



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

import asyncio
import re

import pytest

import svcs

from .fake_factories import async_int_factory, async_str_cleanup_factory, nop
from .ifaces import AnotherService, Interface, Service, YetAnotherService


def test_register_factory_get(registry, container):
    """
    register_factory registers a factory and get returns the service.

    The service is cached.
    """
    registry.register_factory(Service, Service)

    svc = container.get(Service)

    assert isinstance(svc, Service)
    assert svc is container.get(Service)


def test_register_factory_get_abstract(registry, container):
    """
    register_factory registers a factory and get_abstract returns the service.

    The service is cached.
    """
    registry.register_factory(Interface, Service)

    svc = container.get_abstract(Interface)

    assert isinstance(svc, Interface)
    assert svc is container.get(Interface)


def test_register_value_multiple(registry, container):
    """
    register_value registers a service object and get returns as many values as
    are requeste.
    """
    registry.register_value(Service, 1)
    registry.register_value(AnotherService, 2)

    assert [1, 2] == container.get(Service, AnotherService)
    assert [1, 2] == container.get(Service, AnotherService)


def test_get_not_found(container):
    """
    Asking for a service that isn't registered raises a ServiceNotFoundError.
    """
    with pytest.raises(svcs.exceptions.ServiceNotFoundError) as ei:
        container.get(Service)

    assert Service is ei.value.args[0]


def test_passes_container_bc_name(registry, container):
    """
    If the factory takes an argument called `svcs_container`, it is passed on
    instantiation.
    """

    def factory(svcs_container):
        return str(svcs_container.get(int))

    registry.register_value(int, 42)
    registry.register_factory(str, factory)

    assert "42" == container.get(str)


def test_passes_container_bc_annotation(registry, container):
    """
    If the factory takes an argument annotated with svcs.Container, it is
    passed on instantiation.
    """

    def factory(foo: svcs.Container):
        return str(foo.get(int))

    registry.register_value(int, 42)
    registry.register_factory(str, factory)

    assert "42" == container.get(str)


def test_get_pings(registry, container, svc):
    """
    get_pings returns a list of ServicePings.
    """
    registry.register_factory(AnotherService, AnotherService)
    registry.register_value(Service, svc, ping=nop)

    assert [Service] == [ping._svc_type for ping in container.get_pings()]


def test_cleanup_called(registry, container):
    """
    Services that have a cleanup have them called on cleanup.
    """
    cleaned_up = False

    def factory():
        nonlocal cleaned_up
        yield 42
        cleaned_up = True

    registry.register_factory(Service, factory)

    container.get(Service)

    assert not cleaned_up

    container.close()

    assert cleaned_up
    assert not container._instantiated
    assert not container._on_close


def test_close_resilient(container, registry, caplog):
    """
    Failing cleanups are logged and ignored. They do not break the cleanup
    process.
    """

    def factory():
        yield 1
        raise Exception

    cleaned_up = False

    def factory_no_boom():
        nonlocal cleaned_up

        yield 3

        cleaned_up = True

    registry.register_factory(Service, factory)
    registry.register_factory(YetAnotherService, factory_no_boom)

    assert 1 == container.get(Service)
    assert 3 == container.get(YetAnotherService)

    assert not cleaned_up

    container.close()

    assert "tests.ifaces.Service" == caplog.records[0].svcs_service_name
    assert cleaned_up


def test_warns_if_generator_does_not_stop_after_cleanup(registry, container):
    """
    If a generator doesn't stop after cleanup, a warning is emitted.
    """

    def factory():
        yield Service()
        yield 42

    registry.register_factory(Service, factory)

    container.get(Service)

    with pytest.warns(UserWarning) as wi:
        container.close()

    assert (
        "Container clean up for 'tests.ifaces.Service' "
        "didn't stop iterating." == wi.pop().message.args[0]
    )


def test_none_is_a_valid_factory_result(registry, container):
    """
    None is a valid factory result and is cached as such.
    """

    i = 0

    def factory():
        nonlocal i
        i += 1
        yield None

    registry.register_factory(Service, factory)

    assert None is container.get(Service)
    assert None is container.get(Service)
    assert 1 == i


@pytest.mark.parametrize(
    "factory", [async_int_factory, async_str_cleanup_factory]
)
def test_get_on_async_factory_raises_type_error(registry, container, factory):
    """
    get() on an async factory raises a TypeError.
    """

    registry.register_factory(Service, factory)

    with pytest.raises(
        TypeError, match=re.escape("Please use `aget()` for async factories.")
    ):
        container.get(Service)


@pytest.mark.asyncio()
class TestAsync:
    async def test_async_factory(self, registry, container):
        """
        A factory can be async.
        """

        async def factory():
            await asyncio.sleep(0)
            return Service()

        registry.register_factory(Service, factory)

        svc = await container.aget(Service)

        assert isinstance(svc, Service)
        assert svc is (await container.aget(Service))

    async def test_aget_works_with_sync_factory(self, registry, container):
        """
        A synchronous factory does not break aget().
        """
        registry.register_factory(Service, Service)

        assert Service() == (await container.aget(Service))

    async def test_aget_works_with_value(self, registry, container):
        """
        A value instead of a factory does not break aget().
        """
        registry.register_value(Service, 42)
        registry.register_value(AnotherService, 23)

        assert [42, 23] == (await container.aget(Service, AnotherService))
        assert [42, 23] == (await container.aget(Service, AnotherService))

    async def test_aget_abstract_works_with_value(self, registry, container):
        """
        A value instead of a factory does not break aget_abstract().
        """
        registry.register_value(int, 42)
        registry.register_value(str, "42")

        assert [42, "42"] == (await container.aget_abstract(int, str))
        assert [42, "42"] == (await container.aget_abstract(int, str))

    async def test_async_cleanup(self, registry, container):
        """
        Async cleanups are handled by aclose.
        """
        cleaned_up = False

        async def factory():
            nonlocal cleaned_up
            await asyncio.sleep(0)

            yield Service()

            await asyncio.sleep(0)
            cleaned_up = True

        registry.register_factory(Service, factory)

        svc = await container.aget(Service)

        assert 1 == len(container._on_close)
        assert Service() == svc
        assert not cleaned_up

        await container.aclose()

        assert cleaned_up
        assert not container._instantiated
        assert not container._on_close

    @pytest.mark.asyncio()
    async def test_aclose_resilient(self, container, registry, caplog):
        """
        Failing cleanups are logged and ignored. They do not break the
        cleanup process.
        """

        def factory():
            yield 1
            raise Exception

        async def async_factory():
            yield 2
            raise Exception

        cleaned_up = False

        async def factory_no_boom():
            nonlocal cleaned_up

            yield 3

            cleaned_up = True

        registry.register_factory(Service, factory)
        registry.register_factory(AnotherService, async_factory)
        registry.register_factory(YetAnotherService, factory_no_boom)

        assert 1 == container.get(Service)
        assert 2 == await container.aget(AnotherService)
        assert 3 == await container.aget(YetAnotherService)

        assert not cleaned_up

        await container.aclose()

        # Inverse order
        assert (
            "tests.ifaces.AnotherService"
            == caplog.records[0].svcs_service_name
        )
        assert "tests.ifaces.Service" == caplog.records[1].svcs_service_name
        assert cleaned_up
        assert not container._instantiated
        assert not container._on_close

    async def test_warns_if_generator_does_not_stop_after_cleanup(
        self, registry, container
    ):
        """
        If a generator doesn't stop after cleanup, a warning is emitted.
        """

        async def factory():
            yield Service()
            yield 42

        registry.register_factory(Service, factory)

        await container.aget(Service)

        with pytest.warns(UserWarning) as wi:
            await container.aclose()

        assert (
            "Container clean up for 'tests.ifaces.Service' "
            "didn't stop iterating." == wi.pop().message.args[0]
        )

    async def test_aping(self, registry, container):
        """
        Async and sync pings work.
        """
        apinged = pinged = False

        async def aping(svc):
            await asyncio.sleep(0)
            nonlocal apinged
            apinged = True

        def ping(svc):
            nonlocal pinged
            pinged = True

        registry.register_value(Service, Service(), ping=aping)
        registry.register_value(AnotherService, AnotherService(), ping=ping)

        (ap, p) = container.get_pings()

        assert ap.is_async
        assert not p.is_async

        await ap.aping()
        await p.aping()

        assert pinged
        assert apinged

    async def test_none_is_a_valid_factory_result(self, registry, container):
        """
        None is a valid factory result and is cached as such.
        """

        i = 0

        async def factory():
            nonlocal i
            i += 1
            yield None

        registry.register_factory(Service, factory)

        assert None is await container.aget(Service)
        assert None is await container.aget(Service)
        assert 1 == i



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from typing import Generator

from aiohttp.web import Application, Request

import svcs


def factory_with_cleanup() -> Generator[int, None, None]:
    yield 1


reg = svcs.Registry()

app = Application()
app = svcs.aiohttp.init_app(app, registry=reg)
app = svcs.aiohttp.init_app(app, registry=reg, middleware_pos=3)
app = svcs.aiohttp.init_app(app)

svcs.aiohttp.register_value(app, int, 1)
svcs.aiohttp.register_value(app, int, 1, ping=lambda: None)

svcs.aiohttp.register_factory(app, str, str)
svcs.aiohttp.register_factory(app, int, factory_with_cleanup)
svcs.aiohttp.register_value(app, str, str, ping=lambda: None)

a: int
b: str
c: bool
d: tuple
e: object
f: float
g: list
h: dict
i: set
j: bytes

request = Request()


async def func() -> None:
    a, b, c, d, e, f, g, h, i, j = await svcs.aiohttp.aget(
        request, int, str, bool, tuple, object, float, list, dict, set, bytes
    )

    await svcs.aiohttp.aclose_registry(app)


con: svcs.Container = svcs.aiohttp.svcs_from(request)



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

import sys

from doctest import ELLIPSIS

import pytest

from sybil import Sybil
from sybil.parsers import myst, rest
from tests.ifaces import Service

import svcs


markdown_examples = Sybil(
    parsers=[
        myst.DocTestDirectiveParser(optionflags=ELLIPSIS),
        myst.PythonCodeBlockParser(doctest_optionflags=ELLIPSIS),
        myst.SkipParser(),
    ],
    patterns=["*.md"],
)

rest_examples = Sybil(
    parsers=[
        rest.DocTestParser(optionflags=ELLIPSIS),
        rest.PythonCodeBlockParser(),
    ],
    patterns=["*.py"],
)

pytest_collect_file = (markdown_examples + rest_examples).pytest()

collect_ignore = []
try:
    import sphinx  # noqa: F401
except ImportError:
    collect_ignore.extend(["docs"])

if sys.version_info >= (3, 12):
    collect_ignore.extend(["tests/test_pyramid.py"])


@pytest.fixture(name="svc")
def _svc():
    return Service()


@pytest.fixture(name="rs")
def _rs(svc):
    return svcs.RegisteredService(Service, Service, False, False, None)


@pytest.fixture(name="registry")
def _registry():
    return svcs.Registry()


@pytest.fixture(name="container")
def _container(registry):
    return svcs.Container(registry)



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

"""
Ensure our fake factories do what we want them to do.
"""

import sys

from typing import AsyncGenerator, Generator

import pytest

from .fake_factories import (
    async_int_factory,
    async_str_cleanup_factory,
    int_factory,
    nop,
    str_cleanup_factory,
)


def test_nop():
    """
    nop takes any argument and returns always None.
    """
    assert None is nop()
    assert None is nop(1)
    assert None is nop(1, x=2)


def test_int_factory():
    """
    int_factory takes no arguments and returns an int.
    """
    assert isinstance(int_factory(), int)


def test_str_cleanup_factory():
    """
    str_cleanup_factory takes no arguments and returns a generator that yields
    a string.
    """
    gen = str_cleanup_factory()

    assert isinstance(gen, Generator)
    assert isinstance(next(gen), str)

    with pytest.raises(StopIteration):
        next(gen)


@pytest.mark.asyncio()
async def test_async_int_factory():
    """
    async_int_factory takes no arguments and returns an int.
    """
    assert isinstance(await async_int_factory(), int)


@pytest.mark.asyncio()
async def test_async_str_cleanup_factory():
    """
    async_str_cleanup_factory takes no arguments and returns an async generator
    that yields a string.
    """
    gen = async_str_cleanup_factory()

    assert isinstance(gen, AsyncGenerator)
    assert isinstance(await anext(gen), str)

    with pytest.raises(StopAsyncIteration):
        await anext(gen)


if sys.version_info < (3, 10):

    def anext(gen: AsyncGenerator):
        return gen.__anext__()



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import inspect
import logging
import sys
import warnings

from collections.abc import Callable
from contextlib import suppress
from inspect import isasyncgenfunction, isawaitable, iscoroutinefunction
from types import TracebackType
from typing import Any, AsyncGenerator, Awaitable, Generator, TypeVar, overload

import attrs

from .exceptions import ServiceNotFoundError


log = logging.getLogger("svcs")

if sys.version_info < (3, 10):

    def anext(gen: AsyncGenerator) -> Any:
        return gen.__anext__()


def _full_name(obj: object) -> str:
    try:
        return f"{obj.__module__}.{obj.__qualname__}"  # type: ignore[attr-defined]
    except AttributeError:
        return repr(obj)


# Default names where to put the container and registry in integrations.
_KEY_REGISTRY = "svcs_registry"
_KEY_CONTAINER = "svcs_container"


@attrs.frozen
class RegisteredService:
    svc_type: type
    factory: Callable = attrs.field(hash=False)
    takes_container: bool
    is_async: bool
    ping: Callable | None = attrs.field(hash=False)

    @property
    def name(self) -> str:
        return _full_name(self.svc_type)

    def __repr__(self) -> str:
        return (
            f"<RegisteredService(svc_type="
            f"{self.name}, "
            f"{self.factory}, "
            f"takes_container={self.takes_container}, "
            f"has_ping={ self.ping is not None}"
            ")>"
        )


@attrs.frozen
class ServicePing:
    """
    A service health check as returned by :meth:`svcs.Container.get_pings`.

    Attributes:
        name: A fully-qualified name of the service type.

        is_async: Whether the service needs to be pinged using :meth:`aping`.

    See also:
        :ref:`health`
    """

    name: str
    is_async: bool
    _svc_type: type
    _ping: Callable
    _container: Container

    def ping(self) -> None:
        """
        Instantiate the service, schedule its cleanup, and call its ping
        method.
        """
        svc: Any = self._container.get(self._svc_type)
        self._ping(svc)

    async def aping(self) -> None:
        """
        Same as :meth:`ping` but instantiate and/or ping asynchronously, if
        necessary.

        Also works with synchronous services, so in an async application, just
        use this.
        """
        svc: Any = await self._container.aget(self._svc_type)
        if iscoroutinefunction(self._ping):
            await self._ping(svc)
        else:
            self._ping(svc)


@attrs.define
class Registry:
    """
    A central registry of recipes for creating services.

    An instance of this should live as long as your application does.

    Also works as a context manager that runs ``on_registry_close`` callbacks
    on exit:

    .. doctest::

        >>> import svcs
        >>> with svcs.Registry() as reg:
        ...     reg.register_value(
        ...         int, 42,
        ...         on_registry_close=lambda: print("closed!")
        ...     )
        closed!

    ``async with`` is also supported.
    """

    _services: dict[type, RegisteredService] = attrs.Factory(dict)
    _on_close: list[tuple[str, Callable | Awaitable]] = attrs.Factory(list)

    def __repr__(self) -> str:
        return f"<svcs.Registry(num_services={len(self._services)})>"

    def __contains__(self, svc_type: type) -> bool:
        """
        Check whether this registry knows how to create *svc_type*:

        .. doctest::

            >>> reg = svcs.Registry()
            >>> reg.register_value(int, 42)
            >>> int in reg
            True
            >>> str in reg
            False
        """
        return svc_type in self._services

    def __enter__(self) -> Registry:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.close()

    async def __aenter__(self) -> Registry:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.aclose()

    def register_factory(
        self,
        svc_type: type,
        factory: Callable,
        *,
        ping: Callable | None = None,
        on_registry_close: Callable | Awaitable | None = None,
    ) -> None:
        """
        Register *factory* to be used when asked for a *svc_type*.

        Repeated registrations overwrite previous ones, but the
        *on_registry_close* callbacks are run all together when the registry is
        closed.

        Args:
            svc_type: The type of the service to register.

            factory: A callable that is used to instantiated *svc_type* if
                asked. If it's a generator, a cleanup is registered after
                instantiation.

                Can also be an async callable or an async generator.

                If *factory* takes a first argument called ``svcs_container``
                or the first argument (of any name) is annotated as being
                :class:`svcs.Container`, the container instance that is
                instantiating the service is passed into the factory as the
                first positional argument.

            ping: A callable that marks the service as having a health check.

                .. seealso::
                    :meth:`Container.get_pings` and :class:`ServicePing`.

            on_registry_close: A callable that is called when the
                :meth:`svcs.Registry.close()` method is called.

                Can also be an async callable or an
                :class:`collections.abc.Awaitable`, then
                :meth:`svcs.Registry.aclose()` must be called.
        """
        rs = self._register_factory(
            svc_type, factory, ping=ping, on_registry_close=on_registry_close
        )

        log.debug(
            "registered factory %r for service type %s",
            factory,
            rs.name,
            extra={
                "svcs_service_name": rs.name,
                "svcs_factory_name": _full_name(factory),
            },
            stack_info=True,
        )

    def register_value(
        self,
        svc_type: type,
        value: object,
        *,
        ping: Callable | None = None,
        on_registry_close: Callable | Awaitable | None = None,
    ) -> None:
        """
        Syntactic sugar for::

           register_factory(
               svc_type,
               lambda: value,
               ping=ping,
               on_registry_close=on_registry_close
           )
        """
        rs = self._register_factory(
            svc_type,
            lambda: value,
            ping=ping,
            on_registry_close=on_registry_close,
        )

        log.debug(
            "registered value %r for service type %s",
            value,
            rs.name,
            extra={"svcs_service_name": rs.name, "svcs_value": value},
            stack_info=True,
        )

    def _register_factory(
        self,
        svc_type: type,
        factory: Callable,
        ping: Callable | None,
        on_registry_close: Callable | Awaitable | None = None,
    ) -> RegisteredService:
        rs = RegisteredService(
            svc_type,
            factory,
            _takes_container(factory),
            iscoroutinefunction(factory) or isasyncgenfunction(factory),
            ping,
        )
        self._services[svc_type] = rs
        if on_registry_close is not None:
            self._on_close.append((rs.name, on_registry_close))
        return rs

    def get_registered_service_for(self, svc_type: type) -> RegisteredService:
        try:
            return self._services[svc_type]
        except KeyError:
            raise ServiceNotFoundError(svc_type) from None

    def close(self) -> None:
        """
        Clear registrations and run synchronous *on_registry_close* callbacks.

        Async callbacks are *not* awaited and a warning is raised

        Errors are logged at warning level, but otherwise ignored.
        """
        for name, oc in reversed(self._on_close):
            if iscoroutinefunction(oc) or isawaitable(oc):
                warnings.warn(
                    f"Skipped async cleanup for {name!r}. "
                    "Use aclose() instead.",
                    # stacklevel doesn't matter here; it's coming from a
                    # framework.
                    stacklevel=1,
                )
                continue

            try:
                log.debug("closing %r", name)
                oc()  # type: ignore[operator]
                log.debug("closed %r", name)
            except Exception:  # noqa: BLE001
                log.warning(
                    "Registry's on_registry_close callback failed for %r.",
                    name,
                    exc_info=True,
                    extra={"svcs_service_name": name},
                )

        self._services.clear()
        self._on_close.clear()

    async def aclose(self) -> None:
        """
        Clear registrations and run all *on_registry_close* callbacks.

        Errors are logged at warning level, but otherwise ignored.

        Also works with synchronous services, so in an async application, just
        use this.
        """
        for name, oc in reversed(self._on_close):
            try:
                if iscoroutinefunction(oc):
                    oc = oc()  # noqa: PLW2901

                if isawaitable(oc):
                    log.debug("async closing %r", name)
                    await oc
                    log.debug("async closed %r", name)
                else:
                    log.debug("closing %r", name)
                    oc()  # type: ignore[operator]
                    log.debug("closed %r", name)
            except Exception:  # noqa: BLE001, PERF203
                log.warning(
                    "Registry's on_registry_close callback failed for %r.",
                    name,
                    exc_info=True,
                    extra={"svcs_service_name": name},
                )

        self._services.clear()
        self._on_close.clear()


def _takes_container(factory: Callable) -> bool:
    """
    Return True if *factory* takes a svcs.Container as its first argument.
    """
    try:
        sig = inspect.signature(factory)
    except Exception:  # noqa: BLE001
        return False

    if not sig.parameters:
        return False

    if len(sig.parameters) != 1:
        msg = "Factories must take 0 or 1 parameters."
        raise TypeError(msg)

    ((name, p),) = tuple(sig.parameters.items())
    if name == "svcs_container":
        return True

    if (annot := p.annotation) is Container or annot == "svcs.Container":
        return True

    return False


T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")
T7 = TypeVar("T7")
T8 = TypeVar("T8")
T9 = TypeVar("T9")
T10 = TypeVar("T10")


@attrs.define
class Container:
    """
    A per-context container for instantiated services and cleanups.

    The instance of this should live as long as a request or a task.

    Also works as a context manager that runs clean ups on exit:

    .. doctest::

        >>> reg = svcs.Registry()
        >>> def factory() -> str:
        ...     yield "Hello World"
        ...     print("Cleaned up!")
        >>> reg.register_factory(str, factory)

        >>> with svcs.Container(reg) as con:
        ...     _ = con.get(str)
        Cleaned up!

    Attributes:

        registry: The :class:`Registry` instance that this container uses for
           service type lookup.
    """

    registry: Registry
    _instantiated: dict[type, object] = attrs.Factory(dict)
    _on_close: list[tuple[str, Generator | AsyncGenerator]] = attrs.Factory(
        list
    )

    def __repr__(self) -> str:
        return (
            f"<Container(instantiated={len(self._instantiated)}, "
            f"cleanups={len(self._on_close)})>"
        )

    def __contains__(self, svc_type: type) -> bool:
        """
        Check whether this container has a cached instance of *svc_type*.
        """
        return svc_type in self._instantiated

    def __enter__(self) -> Container:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.close()

    async def __aenter__(self) -> Container:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.aclose()

    def forget_about(self, svc_type: type) -> None:
        """
        Forget cached instances of *svc_type* if there are any.
        """
        with suppress(KeyError):
            del self._instantiated[svc_type]

    def close(self) -> None:
        """
        Run all registered *synchronous* cleanups.

        Async closes are *not* awaited and a warning is raised.

        Errors are logged at warning level, but otherwise ignored.
        """
        for name, gen in reversed(self._on_close):
            try:
                if isinstance(gen, AsyncGenerator):
                    warnings.warn(
                        f"Skipped async cleanup for {name!r}. "
                        "Use aclose() instead.",
                        # stacklevel doesn't matter here; it's coming from a framework.
                        stacklevel=1,
                    )
                    continue

                next(gen)

                warnings.warn(
                    f"Container clean up for {name!r} didn't stop iterating.",
                    stacklevel=1,
                )
            except StopIteration:  # noqa: PERF203
                pass
            except Exception:  # noqa: BLE001
                log.warning(
                    "Container clean up failed for %r.",
                    name,
                    exc_info=True,
                    extra={"svcs_service_name": name},
                )

        self._on_close.clear()
        self._instantiated.clear()

    async def aclose(self) -> None:
        """
        Run *all* registered cleanups -- synchronous **and** asynchronous.

        Errors are logged at warning level, but otherwise ignored.

        Also works with synchronous services, so in an async application, just
        use this.
        """
        for name, gen in reversed(self._on_close):
            try:
                if isinstance(gen, AsyncGenerator):
                    await anext(gen)
                else:
                    next(gen)

                warnings.warn(
                    f"Container clean up for {name!r} didn't stop iterating.",
                    stacklevel=1,
                )

            except (StopAsyncIteration, StopIteration):  # noqa: PERF203
                pass
            except Exception:  # noqa: BLE001
                log.warning(
                    "Container clean up failed for %r.",
                    name,
                    exc_info=True,
                    extra={"svcs_service_name": name},
                )

        self._on_close.clear()
        self._instantiated.clear()

    def get_pings(self) -> list[ServicePing]:
        """
        Return all services that have defined a *ping* and bind them to this
        container.

        Returns:
            A sequence of services that have registered a ping callable.
        """
        return [
            ServicePing(
                rs.name,
                rs.is_async or iscoroutinefunction(rs.ping),
                rs.svc_type,
                rs.ping,
                self,
            )
            for rs in self.registry._services.values()
            if rs.ping is not None
        ]

    def get_abstract(self, *svc_types: type) -> Any:
        """
        Like :meth:`get` but is annotated to return :data:`typing.Any` which
        allows it to be used with abstract types like :class:`typing.Protocol`
        or :mod:`abc` classes.

        Note:
             See :doc:`typing-caveats` why this is necessary.
        """
        return self.get(*svc_types)

    async def aget_abstract(self, *svc_types: type) -> Any:
        """
        Same as :meth:`get_abstract` but instantiates asynchronously, if
        necessary.

        Also works with synchronous services, so in an async application, just
        use this.
        """
        return await self.aget(*svc_types)

    @overload
    def get(self, svc_type: type[T1], /) -> T1:
        ...

    @overload
    def get(
        self, svc_type1: type[T1], svc_type2: type[T2], /
    ) -> tuple[T1, T2]:
        ...

    @overload
    def get(
        self, svc_type1: type[T1], svc_type2: type[T2], svc_type3: type[T3], /
    ) -> tuple[T1, T2, T3]:
        ...

    @overload
    def get(
        self,
        svc_type1: type[T1],
        svc_type2: type[T2],
        svc_type3: type[T3],
        svc_type4: type[T4],
        /,
    ) -> tuple[T1, T2, T3, T4]:
        ...

    @overload
    def get(
        self,
        svc_type1: type[T1],
        svc_type2: type[T2],
        svc_type3: type[T3],
        svc_type4: type[T4],
        svc_type5: type[T5],
        /,
    ) -> tuple[T1, T2, T3, T4, T5]:
        ...

    @overload
    def get(
        self,
        svc_type1: type[T1],
        svc_type2: type[T2],
        svc_type3: type[T3],
        svc_type4: type[T4],
        svc_type5: type[T5],
        svc_type6: type[T6],
        /,
    ) -> tuple[T1, T2, T3, T4, T5, T6]:
        ...

    @overload
    def get(
        self,
        svc_type1: type[T1],
        svc_type2: type[T2],
        svc_type3: type[T3],
        svc_type4: type[T4],
        svc_type5: type[T5],
        svc_type6: type[T6],
        svc_type7: type[T7],
        /,
    ) -> tuple[T1, T2, T3, T4, T5, T6, T7]:
        ...

    @overload
    def get(
        self,
        svc_type1: type[T1],
        svc_type2: type[T2],
        svc_type3: type[T3],
        svc_type4: type[T4],
        svc_type5: type[T5],
        svc_type6: type[T6],
        svc_type7: type[T7],
        svc_type8: type[T8],
        /,
    ) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8]:
        ...

    @overload
    def get(
        self,
        svc_type1: type[T1],
        svc_type2: type[T2],
        svc_type3: type[T3],
        svc_type4: type[T4],
        svc_type5: type[T5],
        svc_type6: type[T6],
        svc_type7: type[T7],
        svc_type8: type[T8],
        svc_type9: type[T9],
        /,
    ) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8, T9]:
        ...

    @overload
    def get(
        self,
        svc_type1: type[T1],
        svc_type2: type[T2],
        svc_type3: type[T3],
        svc_type4: type[T4],
        svc_type5: type[T5],
        svc_type6: type[T6],
        svc_type7: type[T7],
        svc_type8: type[T8],
        svc_type9: type[T9],
        svc_type10: type[T10],
        /,
    ) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10]:
        ...

    def get(self, *svc_types: type) -> object:
        """
        Get services of *svc_types*.

        Instantiate them if necessary and register their cleanup.

        Returns:
             ``svc_types[0]`` | ``tuple[*svc_types]``: If one service is
             requested, it's returned directly. If multiple are requested, a
             tuple of services is returned.
        """
        rv = []
        for svc_type in svc_types:
            if (
                svc := self._instantiated.get(svc_type, attrs.NOTHING)
            ) is not attrs.NOTHING:
                rv.append(svc)
                continue

            rs = self.registry.get_registered_service_for(svc_type)
            if rs.is_async:
                msg = "Please use `aget()` for async factories."
                raise TypeError(msg)

            svc = rs.factory(self) if rs.takes_container else rs.factory()

            if isinstance(svc, Generator):
                self._on_close.append((rs.name, svc))
                svc = next(svc)

            self._instantiated[svc_type] = svc

            rv.append(svc)

        if len(rv) == 1:
            return rv[0]

        return rv

    @overload
    async def aget(self, svc_type: type[T1], /) -> T1:
        ...

    @overload
    async def aget(
        self, svc_type1: type[T1], svc_type2: type[T2], /
    ) -> tuple[T1, T2]:
        ...

    @overload
    async def aget(
        self, svc_type1: type[T1], svc_type2: type[T2], svc_type3: type[T3], /
    ) -> tuple[T1, T2, T3]:
        ...

    @overload
    async def aget(
        self,
        svc_type1: type[T1],
        svc_type2: type[T2],
        svc_type3: type[T3],
        svc_type4: type[T4],
        /,
    ) -> tuple[T1, T2, T3, T4]:
        ...

    @overload
    async def aget(
        self,
        svc_type1: type[T1],
        svc_type2: type[T2],
        svc_type3: type[T3],
        svc_type4: type[T4],
        svc_type5: type[T5],
        /,
    ) -> tuple[T1, T2, T3, T4, T5]:
        ...

    @overload
    async def aget(
        self,
        svc_type1: type[T1],
        svc_type2: type[T2],
        svc_type3: type[T3],
        svc_type4: type[T4],
        svc_type5: type[T5],
        svc_type6: type[T6],
        /,
    ) -> tuple[T1, T2, T3, T4, T5, T6]:
        ...

    @overload
    async def aget(
        self,
        svc_type1: type[T1],
        svc_type2: type[T2],
        svc_type3: type[T3],
        svc_type4: type[T4],
        svc_type5: type[T5],
        svc_type6: type[T6],
        svc_type7: type[T7],
        /,
    ) -> tuple[T1, T2, T3, T4, T5, T6, T7]:
        ...

    @overload
    async def aget(
        self,
        svc_type1: type[T1],
        svc_type2: type[T2],
        svc_type3: type[T3],
        svc_type4: type[T4],
        svc_type5: type[T5],
        svc_type6: type[T6],
        svc_type7: type[T7],
        svc_type8: type[T8],
        /,
    ) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8]:
        ...

    @overload
    async def aget(
        self,
        svc_type1: type[T1],
        svc_type2: type[T2],
        svc_type3: type[T3],
        svc_type4: type[T4],
        svc_type5: type[T5],
        svc_type6: type[T6],
        svc_type7: type[T7],
        svc_type8: type[T8],
        svc_type9: type[T9],
        /,
    ) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8, T9]:
        ...

    @overload
    async def aget(
        self,
        svc_type1: type[T1],
        svc_type2: type[T2],
        svc_type3: type[T3],
        svc_type4: type[T4],
        svc_type5: type[T5],
        svc_type6: type[T6],
        svc_type7: type[T7],
        svc_type8: type[T8],
        svc_type9: type[T9],
        svc_type10: type[T10],
        /,
    ) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10]:
        ...

    async def aget(self, *svc_types: type) -> object:
        """
        Same as :meth:`get` but instantiates asynchronously, if necessary.

        Also works with synchronous services, so in an async application, just
        use this.
        """
        rv = []
        for svc_type in svc_types:
            if (
                svc := self._instantiated.get(svc_type, attrs.NOTHING)
            ) is not attrs.NOTHING:
                rv.append(svc)
                continue

            rs = self.registry.get_registered_service_for(svc_type)
            svc = rs.factory()

            if isinstance(svc, AsyncGenerator):
                self._on_close.append((rs.name, svc))
                svc = await anext(svc)
            elif isawaitable(svc):
                svc = await svc

            self._instantiated[rs.svc_type] = svc

            rv.append(svc)

        if len(rv) == 1:
            return rv[0]

        return rv



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from unittest.mock import Mock

import pytest

from .ifaces import AnotherService, Service


class TestContainer:
    def test_get_pings_empty(self, container):
        """
        get_pings returns an empty list if there are no pings.
        """
        assert [] == container.get_pings()

    def test_forget_about_nothing_registered(self, container):
        """
        forget_about does nothing if nothing has been registered.
        """
        container.forget_about(Service)

    def test_forget_about_no_cleanup(self, container, rs, svc):
        """
        forget_about removes the registered service from the container.
        """
        container._instantiated[rs.svc_type] = (rs, svc)

        container.forget_about(Service)

        assert {} == container._instantiated
        assert [] == container._on_close

    @pytest.mark.asyncio()
    async def test_repr(self, registry, container):
        """
        The repr counts correctly.
        """

        def factory():
            yield 42

        async def async_factory():
            yield 42

        registry.register_factory(Service, factory)
        registry.register_factory(AnotherService, async_factory)

        container.get(Service)
        await container.aget(AnotherService)

        assert "<Container(instantiated=2, cleanups=2)>" == repr(container)

    def test_contains(self, container):
        """
        If a service is instantiated within a container, `in` returns True,
        False otherwise.
        """
        container.registry.register_value(int, 42)

        assert int not in container

        container.get(int)

        assert int in container

    def test_context_manager(self, container):
        """
        The container is also a context manager that closes on exit.
        """
        closed = False

        def factory():
            yield 42
            nonlocal closed
            closed = True

        container.registry.register_factory(int, factory)

        with container:
            assert 42 == container.get(int)

        assert closed

    @pytest.mark.asyncio()
    async def test_async_context_manager(self, container):
        """
        The container is also an async context manager that acloses on exit.
        """
        closed = False

        async def factory():
            yield 42
            nonlocal closed
            closed = True

        container.registry.register_factory(int, factory)

        async with container:
            assert 42 == await container.aget(int)

        assert closed


class TestServicePing:
    def test_ping(self, registry, container):
        """
        Calling ping instantiates the service using its factory, appends it to
        the cleanup list, and calls the service's ping method.
        """

        cleaned_up = False

        def factory():
            nonlocal cleaned_up
            yield Service()
            cleaned_up = True

        ping = Mock(spec_set=["__call__"])
        registry.register_factory(Service, factory, ping=ping)

        (svc_ping,) = container.get_pings()

        svc_ping.ping()

        ping.assert_called_once()

        assert not cleaned_up

        container.close()

        assert cleaned_up
        assert not container._instantiated
        assert not container._on_close



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Callable
from typing import overload

from aiohttp import web

import svcs

from ._core import (
    _KEY_CONTAINER,
    _KEY_REGISTRY,
    T1,
    T2,
    T3,
    T4,
    T5,
    T6,
    T7,
    T8,
    T9,
    T10,
)


def svcs_from(request: web.Request) -> svcs.Container:
    """
    Get the current container from *request*.
    """
    return request[_KEY_CONTAINER]  # type: ignore[no-any-return]


def init_app(
    app: web.Application,
    *,
    registry: svcs.Registry | None = None,
    middleware_pos: int = 0,
) -> web.Application:
    """
    Initialize the application.

    Inserts the *svcs* middleware at *middleware_pos* which is 0 by default, so
    you can use :func:`svcs_from` and :func:`aget` in other middlewares.
    """
    app[_KEY_REGISTRY] = registry or svcs.Registry()
    app.middlewares.insert(middleware_pos, svcs_middleware)
    app.on_cleanup.append(aclose_registry)

    return app


@web.middleware
async def svcs_middleware(
    request: web.Request, handler: Callable
) -> web.Response:
    async with svcs.Container(request.app[_KEY_REGISTRY]) as container:
        request[_KEY_CONTAINER] = container

        return await handler(request)  # type: ignore[no-any-return]


def register_value(
    app: web.Application,
    svc_type: type,
    value: object,
    *,
    ping: Callable | None = None,
    on_registry_close: Callable | None = None,
) -> None:
    """
    Same as :meth:`svcs.Registry.register_value()`, but uses registry on
    *app*.
    """
    app[_KEY_REGISTRY].register_value(
        svc_type, value, ping=ping, on_registry_close=on_registry_close
    )


def register_factory(
    app: web.Application,
    svc_type: type,
    factory: Callable,
    *,
    ping: Callable | None = None,
    on_registry_close: Callable | None = None,
) -> None:
    """
    Same as :meth:`svcs.Registry.register_factory()`, but uses registry on
    *app*.
    """
    app[_KEY_REGISTRY].register_factory(
        svc_type, factory, ping=ping, on_registry_close=on_registry_close
    )


def get_registry(app: web.Application) -> svcs.Registry:
    """
    Get the registry from *app*.
    """
    return app[_KEY_REGISTRY]  # type: ignore[no-any-return]


async def aclose_registry(app: web.Application) -> None:
    """
    Close the registry on *app*, if present.

    You probably don't have to call this yourself, because it's registered for
    the application as an {attr}`aiohttp.web.Application.on_cleanup` callback.

    .. seealso:: :ref:`aiohttp-cleanup`
    """
    if reg := app.get(_KEY_REGISTRY):
        await reg.aclose()


def get_pings(request: web.Request) -> list[svcs.ServicePing]:
    """
    Same as :meth:`svcs.Container.get_pings`, but uses the container from
    *request*.

    .. seealso:: :ref:`aiohttp-health`
    """
    return request[_KEY_CONTAINER].get_pings()  # type: ignore[no-any-return]


@overload
async def aget(request: web.Request, svc_type: type[T1], /) -> T1:
    ...


@overload
async def aget(
    request: web.Request, svc_type1: type[T1], svc_type2: type[T2], /
) -> tuple[T1, T2]:
    ...


@overload
async def aget(
    request: web.Request,
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    /,
) -> tuple[T1, T2, T3]:
    ...


@overload
async def aget(
    request: web.Request,
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    /,
) -> tuple[T1, T2, T3, T4]:
    ...


@overload
async def aget(
    request: web.Request,
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    /,
) -> tuple[T1, T2, T3, T4, T5]:
    ...


@overload
async def aget(
    request: web.Request,
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6]:
    ...


@overload
async def aget(
    request: web.Request,
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    svc_type7: type[T7],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6, T7]:
    ...


@overload
async def aget(
    request: web.Request,
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    svc_type7: type[T7],
    svc_type8: type[T8],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8]:
    ...


@overload
async def aget(
    request: web.Request,
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    svc_type7: type[T7],
    svc_type8: type[T8],
    svc_type9: type[T9],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8, T9]:
    ...


@overload
async def aget(
    request: web.Request,
    svc_type1: type[T1],
    svc_type2: type[T2],
    svc_type3: type[T3],
    svc_type4: type[T4],
    svc_type5: type[T5],
    svc_type6: type[T6],
    svc_type7: type[T7],
    svc_type8: type[T8],
    svc_type9: type[T9],
    svc_type10: type[T10],
    /,
) -> tuple[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10]:
    ...


async def aget(request: web.Request, *svc_types: type) -> object:
    """
    Same as :meth:`svcs.Container.aget`, but uses the container from *request*.
    """
    return await request[_KEY_CONTAINER].aget(*svc_types)



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from unittest.mock import Mock

import pytest

import svcs

from tests.fake_factories import nop
from tests.ifaces import AnotherService, Interface, Service


try:
    import flask

    from svcs.flask import teardown
except ImportError:
    pytest.skip("Flask not installed", allow_module_level=True)


@pytest.fixture(name="app")
def _app():
    return flask.Flask("tests")


@pytest.fixture(name="clean_app_ctx")
def _clean_app_ctx(registry, app):
    svcs.flask.init_app(app, registry=registry)
    with app.app_context() as ctx:
        yield ctx


@pytest.fixture(name="container")
def _container(clean_app_ctx):
    return svcs.flask.svcs_from()


@pytest.mark.usefixtures("clean_app_ctx")
class TestFlask:
    def test_register_value_multiple(self, registry):
        """
        register_value registers a service object on an app and get returns as
        many values as are requeste.
        """
        registry.register_value(Service, 1)
        registry.register_value(AnotherService, 2)

        assert [1, 2] == svcs.flask.get(Service, AnotherService)
        assert [1, 2] == svcs.flask.get(Service, AnotherService)

    def test_cleanup_added(self, registry):
        """
        get() handles the case where there is already a cleanup registered.
        """

        cleanup1 = Mock()
        cleanup2 = Mock()

        def factory1():
            yield Service()
            cleanup1()

        def factory2():
            yield AnotherService()
            cleanup2()

        registry.register_factory(Service, factory1)
        svcs.flask.replace_factory(AnotherService, factory2)

        svc1 = svcs.flask.get(Service)
        svc2 = svcs.flask.get(AnotherService)

        assert isinstance(svc1, Service)
        assert isinstance(svc2, AnotherService)
        assert 2 == len(flask.g.svcs_container._on_close)

        teardown(None)

        cleanup1.assert_called_once_with()
        cleanup2.assert_called_once_with()

    def test_overwrite_value(self, registry):
        """
        It's possible to overwrite an already registered type.
        """
        registry.register_value(Interface, Service(), ping=nop)

        assert isinstance(svcs.flask.get(Interface), Interface)

        svcs.flask.replace_value(Interface, AnotherService())

        assert isinstance(svcs.flask.get(Interface), AnotherService)
        assert [] == svcs.flask.get_pings()

    def test_overwrite_factory(self):
        """
        It's possible to overwrite an already registered type using a factory.
        """
        svcs.flask.replace_value(Interface, Service(), ping=nop)

        assert isinstance(svcs.flask.get(Interface), Interface)

        svcs.flask.replace_factory(Interface, AnotherService)

        assert isinstance(svcs.flask.get(Interface), AnotherService)
        assert [] == svcs.flask.get_pings()

    def test_cache(self):
        """
        Getting a service twice within the same request returns the same
        object.
        """
        svcs.flask.replace_factory(Interface, Service)

        assert svcs.flask.get(Interface) is svcs.flask.get(Interface)

    def test_not_found(self):
        """
        Trying to get a service that hasn't been registered raises a
        ServiceNotFoundError.
        """
        with pytest.raises(svcs.exceptions.ServiceNotFoundError):
            svcs.flask.get(Interface)

    def test_get_pingeable(self):
        """
        get_pingable returns only pingable svcs.
        """
        svcs.flask.replace_factory(Service, Service)
        svcs.flask.replace_factory(AnotherService, AnotherService, ping=nop)

        assert [AnotherService] == [
            ping._svc_type for ping in svcs.flask.get_pings()
        ]

    def test_cleanup_purge_tolerant(self, container):
        """
        If other svcs are registered, they are ignored by the cleanup
        purge.
        """

        def factory1():
            yield Service()

        def factory2():
            yield AnotherService()

        svcs.flask.replace_factory(Interface, factory1)
        svcs.flask.replace_factory(AnotherService, factory2)

        svcs.flask.get(Interface)
        svcs.flask.get(AnotherService)

        assert 2 == len(container._on_close)

        svcs.flask.replace_factory(Interface, Service)

        svcs.flask.get(Interface)
        svcs.flask.get(AnotherService)

        assert 2 == len(container._on_close)

    @pytest.mark.asyncio()
    async def test_teardown_warns_on_async_on_close(self, container):
        """
        teardown() warns if there are async cleanups.
        """

        async def factory():
            yield Service()

        svcs.flask.replace_factory(Service, factory)

        await container.aget(Service)

        with pytest.warns(UserWarning) as wi:
            teardown(None)

        w = wi.pop()

        assert 0 == len(wi.list)
        assert (
            "Skipped async cleanup for 'tests.ifaces.Service'. "
            "Use aclose() instead." == w.message.args[0]
        )

    def test_register_factory_get_abstract(self, registry, container):
        """
        register_factory registers a factory and get_abstract returns the service.

        The service is cached.
        """
        registry.register_factory(Interface, Service)

        svc = container.get_abstract(Interface)

        assert isinstance(svc, Interface)
        assert svc is svcs.flask.get_abstract(Interface)

    def test_svcs_from(self, container):
        """
        svcs_from() returns the container the same container as that is on g.
        """
        assert (
            container
            is svcs.flask.svcs_from()
            is flask.g.svcs_container
            is svcs.flask.svcs_from()
        )


class TestNonContextHelpers:
    def test_register_factory_helper(self, registry, app):
        """
        register_factory() registers a factory to the app that is passed.
        """
        svcs.flask.init_app(app, registry=registry)

        svcs.flask.register_factory(app, Interface, Service)

        assert Interface in registry._services

    def test_register_value_helper(self, registry, app):
        """
        register_value() registers a value to the app that is passed.
        """
        svcs.flask.init_app(app, registry=registry)

        svcs.flask.register_value(app, Interface, 42)

        assert Interface in registry._services


class TestInitApp:
    def test_implicit_registry(self):
        """
        init_app() creates a registry if one isn't provided.
        """
        app = flask.Flask("tests")
        svcs.flask.init_app(app)

        assert isinstance(app.config["svcs_registry"], svcs.Registry)

    def test_explicit_registry(self):
        """
        If a registry is passed to init_app(), it's used.
        """
        registry = svcs.Registry()
        app = flask.Flask("tests")
        svcs.flask.init_app(app, registry=registry)

        assert registry is app.config["svcs_registry"]


class TestCloseRegistry:
    def test_nop(self):
        """
        close_registry() does nothing if there's no registry in app.
        """
        app = flask.Flask("tests")
        svcs.flask.close_registry(app)

    def test_closes(self, app):
        """
        close_registry() runs the registry's close() method.
        """
        close = Mock()

        svcs.flask.init_app(app)

        svcs.flask.register_factory(
            app, Interface, Service, on_registry_close=close
        )

        svcs.flask.close_registry(app)

        assert close.called



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from typing import Generator

import pyramid

from pyramid.config import Configurator

import svcs


def factory_with_cleanup() -> Generator[int, None, None]:
    yield 1


config = Configurator(settings={})

svcs.pyramid.init(config)

svcs.pyramid.register_value(config, int, 1)
svcs.pyramid.register_value(config, int, 1, ping=lambda: None)

svcs.pyramid.register_factory(config, str, str)
svcs.pyramid.register_factory(config, int, factory_with_cleanup)
svcs.pyramid.register_value(config, str, str, ping=lambda: None)

o1: object = svcs.pyramid.get(object)
o2: int = svcs.pyramid.get_abstract(object)

a: int
b: str
c: bool
d: tuple
e: object
f: float
g: list
h: dict
i: set
j: bytes
a, b, c, d, e, f, g, h, i, j = svcs.pyramid.get(
    int, str, bool, tuple, object, float, list, dict, set, bytes
)

reg: svcs.Registry = svcs.pyramid.get_registry(config)
reg = svcs.pyramid.get_registry()

con: svcs.Container = svcs.pyramid.svcs_from()
con = svcs.pyramid.svcs_from(pyramid.request.Request())

svcs.pyramid.close_registry(config)



from __future__ import annotations

import json

from pyramid.request import Request
from pyramid.response import Response
from pyramid.view import view_config

from svcs.pyramid import svcs_from


@view_config(route_name="healthy")
def healthy_view(request: Request) -> Response:
    ok: list[str] = []
    failing: list[dict[str, str]] = []
    status = 200

    for svc in svcs_from(request).get_pings():
        try:
            svc.ping()
            ok.append(svc.name)
        except Exception as e:
            failing.append({svc.name: repr(e)})
            status = 500

    return Response(
        content_type="application/json",
        status=status,
        body=json.dumps({"ok": ok, "failing": failing}).encode("ascii"),
    )



# SPDX-FileCopyrightText: 2023 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from unittest.mock import Mock

import pytest

import svcs

from tests.fake_factories import int_factory
from tests.ifaces import AnotherService, Service


try:
    import webtest

    from pyramid.config import Configurator
    from pyramid.view import view_config
except ImportError:
    pytest.skip("Pyramid not installed", allow_module_level=True)


def make_config(view=None):
    config = Configurator(settings={})
    svcs.pyramid.init(config)

    if view is not None:
        config.add_route(view, "/")
    config.scan()

    return config


def make_test_app(view=None):
    config = make_config(view)

    return webtest.TestApp(config.make_wsgi_app())


@pytest.fixture(name="config")
def _config():
    return make_config()


def test_close_nop():
    """
    Closing a config that has no svcs_registry does nothing.
    """
    svcs.pyramid.close_registry(Mock(registry={}))


def test_close(config):
    """
    Closing a config with svcs_registry calls on_registry_close callbacks on
    the registered svcs.pyramid.
    """
    orc = Mock()

    svcs.pyramid.register_factory(
        config, int, int_factory, on_registry_close=orc
    )

    svcs.pyramid.close_registry(config)

    assert orc.called


@view_config(route_name="tl_view", renderer="json")
def tl_view(request):
    """
    Thread locals return the same objects as the direct way.
    """
    svc = svcs.pyramid.get(Service)
    svcs.pyramid.get(float)

    assert (
        svc
        is svcs.pyramid.svcs_from(request).get(Service)
        is svcs.pyramid.get_abstract(Service)
    )
    assert (
        request.registry["svcs_registry"]
        is svcs.pyramid.get_registry()
        is svcs.pyramid.get_registry(request)
    )
    assert (
        request.svcs_container
        is svcs.pyramid.svcs_from()
        is svcs.pyramid.svcs_from(request)
    )

    return {"svc": svc}


def test_thread_locals():
    """
    Thread locals are available in views.
    """
    tapp = make_test_app("tl_view")

    closed = False

    def closing_factory():
        yield 1.0

        nonlocal closed
        closed = True

    svcs.pyramid.get_registry(tapp.app).register_value(Service, 42)
    svcs.pyramid.register_value(tapp.app, AnotherService, 23)
    svcs.pyramid.register_factory(tapp.app, float, closing_factory)

    assert {"svc": 42} == tapp.get("/").json
    assert closed
